/*
 * Accept.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Accept.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/PDA/DPDA.h>

#include <deque>

namespace automaton {

namespace run {

bool Accept::accept(const automaton::Automaton& automaton, const string::LinearString& string) {
	std::pair<const string::LinearString*, bool> data(&string, false);
	automaton.getData().Accept((void*) &data, Accept::ACCEPT);
	return data.second;
}

bool Accept::accept(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) return false;
		state = next->second;
	}
	return automaton.getFinalStates().count(state);
}

bool recursiveAccept(const automaton::NFA& automaton, const automaton::State& state, const string::LinearString& string) {
	if( string.getContent().empty( ) )
		return automaton.getFinalStates().count( state );

	const alphabet::Symbol & inputSymbol = string.getContent().front( );

	string::LinearString newString(std::vector<alphabet::Symbol>{ string.getContent().begin() + 1, string.getContent().end()});

	for( const auto & transition : automaton.getTransitionsFromState( state ) )
		for(const auto& to : transition.second)
			if( transition.first.second == inputSymbol && recursiveAccept( automaton, to, newString ) )
				return true;

	return false;
}

bool Accept::accept( const automaton::NFA& automaton, const string::LinearString & string ) {
	return recursiveAccept(automaton, automaton.getInitialState(), string);
}

bool Accept::accept(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		int pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			int popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(int i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return false;
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);
	}
	if(automaton.getFinalStates().empty()) {
		return pushdownStore.empty();
	} else {
		return automaton.getFinalStates().count(state);
	}
}

void Accept::Visit(void*, const automaton::EpsilonNFA& ) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void Accept::Visit(void*, const automaton::MultiInitialStateNFA& ) const {
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void Accept::Visit(void* data, const automaton::NFA& automaton) const {
	std::pair<const string::LinearString*, bool> & res = *((std::pair<const string::LinearString*, bool>*) data);
	res.second = this->accept(automaton, *res.first);
}

void Accept::Visit(void* data, const automaton::DFA& automaton) const {
	std::pair<const string::LinearString*, bool> & res = *((std::pair<const string::LinearString*, bool>*) data);
	res.second = this->accept(automaton, *res.first);
}

void Accept::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void Accept::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void Accept::Visit(void* data, const DPDA& automaton) const {
	std::pair<const string::LinearString*, bool> & res = *((std::pair<const string::LinearString*, bool>*) data);
	res.second = this->accept(automaton, *res.first);
}

void Accept::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void Accept::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void Accept::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void Accept::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void Accept::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void Accept::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void Accept::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void Accept::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void Accept::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const Accept Accept::ACCEPT;

} /* namespace run */

} /* namespace automaton */
